
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <regex.h>
#include <unistd.h>

static unsigned read_next_address(FILE * trace_in, regex_t * prp)
{
    regmatch_t pmatch[2];
    char buf[BUFSIZ];

    if (!fgets(buf, BUFSIZ - 1, trace_in)) {
        return 0;
    }
    if (regexec(prp, buf, 2, pmatch, 0) == 0)  {
        assert(pmatch[1].rm_so > 0);
        buf[pmatch[1].rm_eo] = '\0';
        return strtol(&buf[pmatch[1].rm_so], NULL, 16);
    }
    fprintf(stderr, "Unable to decode trace input:\n%s\n", buf);
    exit(1);
}


int main(int argc, char ** argv)
{
    if (argc != 4) {
        printf("usage: trunc <trace_in file> <nm file> <trace_out file>\n");
        return 1;
    }
    const char * trace_in_file = argv[1];
    const char * nm_file = argv[2];
    const char * trace_out_file = argv[3];

    FILE * trace_in = fopen(trace_in_file, "rt");
    if (trace_in == NULL) {
        printf("unable to open trace_in file\n");
        return 1;
    }

    FILE * nm_in = fopen(nm_file, "rt");
    if (nm_in == NULL) {
        printf("unable to open nm_in file\n");
        return 1;
    }

    FILE * trace_out = fopen(trace_out_file, "wb");
    if (trace_out == NULL) {
        printf("unable to open trace_out file\n");
        return 1;
    }

    unsigned exit_at = 0;
    unsigned main_at = 0;

    // scan nm file to identify functions
    char buf[BUFSIZ];
    regex_t pr;
    regmatch_t pmatch[4];

    int rc = regcomp(&pr, 
            "^([0-9a-f]{8}) ([0-9a-f]{8}|) *[tT] ([^ ]*).$", REG_EXTENDED);
    assert(rc == 0);

    while (fgets(buf, BUFSIZ - 1, nm_in)) {
        if (regexec(&pr, buf, 4, pmatch, 0) == 0) {
            assert(pmatch[1].rm_so == 0);
            assert(pmatch[2].rm_so > 0);
            assert(pmatch[3].rm_so > 0);

            buf[pmatch[1].rm_eo] = '\0';
            unsigned start = strtol(&buf[pmatch[1].rm_so], NULL, 16);

            unsigned size = 0;
            if (pmatch[2].rm_so != pmatch[2].rm_eo) {
                buf[pmatch[2].rm_eo] = '\0';
                size = strtol(&buf[pmatch[2].rm_so], NULL, 16);
            }
            buf[pmatch[3].rm_eo] = '\0';
            const char * name = &buf[pmatch[3].rm_so];

            if (strcmp(name, "exit") == 0) {
                exit_at = start;
                printf("exit at %x\n", exit_at);
            } else if (strcmp(name, "main") == 0) {
                main_at = start;
                printf("main at %x\n", main_at);
            }
        }
    }
    regfree(&pr);
    assert (exit_at);
    assert (main_at);

    rc = regcomp(&pr, 
            "^Trace 0x[0-9a-f]{8} .([0-9a-f]{8})..*$", REG_EXTENDED);
    assert(rc == 0);

    unsigned prev_address = 0;
    unsigned address = 0;
    unsigned i = 0;

    do {
        prev_address = address;
        address = read_next_address(trace_in, &pr);
        i++;
        assert(address);
    } while (address != main_at);

    // prev_address == call site for main
    unsigned return_site = prev_address + 4;

    printf("return site at %x\n", return_site);
    printf("prologue size %u\n", i);

    i = 0;

    do {
        size_t rc = fwrite(&address, 4, 1, trace_out);
        assert(rc == 1);

        address = read_next_address(trace_in, &pr);
        i++;
        assert(address);
        assert(address != exit_at);
    } while (address != return_site);

    printf("run size %u\n", i);
    printf("end at %x (not in file)\n", address);

    regfree(&pr);
    fclose(trace_out);
    fclose(trace_in);
    fclose(nm_in);
    return 0;
}


