#include "rtos.h"


void dump(unsigned b)
{
    unsigned i;
    unsigned * m;
    b *= 16;
    b += 0x10000000;

    m = (unsigned *) b;
    for (i = 0; i < 4; i++) {
        printhex(m[i]);
        print(" ");
    }
    print("\r\n");
}

void inspect_sp(void)
{
    unsigned i, j;
    i = 7;
    print("stack pointer: ");
    asm volatile(
        "put %1, rfsl0\n"
        "get %0, rfsl0\n" : "=r"(j) : "r"((i << 28)) );
    printhex(j);
    print("\n");
}


void inspection(void)
{
    unsigned i, j;

    print("registers...\n");
    for(i = 0; i < 16; i++) {
        asm volatile(
            "put %1, rfsl0\n"
            "get %0, rfsl0\n" : "=r"(j) : "r"((i << 28)) );
        printhex(j);
        if ((i % 4) == 3) {
            print("\n");
        } else {
            print(" ");
        }
    }
    while(1) {}
}

void mtest(void)
{
    unsigned i, j = 0;
    unsigned * ext_mem_loc = (unsigned *) 0x91000000U;
    unsigned tmax = 0;
    unsigned * m;
    print("hello\r\n");
    dump(0);
    set_trigger(100000);

#if 0
    asm volatile("msrset r18, 0xa0"); /* IC, DC on */

    for (i = 0; i < 2000; i ++) {
        unsigned a, b, t0, t1, t2;

        a = get_time();
        asm volatile(
            "addi r18, r0, 0x90000000\n");
        b = get_time();
        t0 = b - a;

        asm volatile(
            "addi r18, r0, 16\n"
            "put r18, rfsl0\n"
            "get %0, rfsl0\n" : "=r"(j) );

        a = get_time();
        asm volatile(
            "addi r18, r0, 0x90000000\n"
            "cput r18, rfsl0\n");
        b = get_time();
        asm volatile(
            "get %0, rfsl0\n" : "=r"(j) );
        t1 = b - a;

        asm volatile(
            "get %0, rfsl0\n" : "=r"(j) );
        
        a = get_time();
        asm volatile(
            "addi r18, r0, 0x90000010\n"
            "cput r18, rfsl0\n"
            "get %0, rfsl0\n" : "=r"(j) );
        b = get_time();
        t2 = b - a;
        t2 -= t0;
       
        if (t2 > tmax) {
            tmax = t2;
        }
        if (i == 10) {
            tmax = 0;
        }
    }
    print("SPM timing - ");
    printhex(tmax);
    print("\n");
    tmax = 0;

    /* Load block 8 */
    asm volatile(
        "addi r18, r0, 8*16\n"
        "put r18, rfsl0\n"
        "addi r18, r0, 0x90000100\n"
        "cput r18, rfsl0\n"
        "get r18, rfsl0\n");
    asm volatile(
        "addi r18, r0, 9*16\n"
        "put r18, rfsl0\n"
        "addi r18, r0, 0x90000110\n"
        "cput r18, rfsl0\n"
        "get r18, rfsl0\n");

    dump(7);
    dump(8);
    dump(9);
    print("(a)\n");

    /* Load something else over block 8 - also Save*/
    asm volatile(
        "addi r18, r0, (8*16)|1\n"
        "put r18, rfsl0\n"
        "addi r18, r0, 0x90000200\n"
        "cput r18, rfsl0\n"
        "get r18, rfsl0\n");
    asm volatile(
        "addi r18, r0, (9*16)|1\n"
        "put r18, rfsl0\n"
        "addi r18, r0, 0x90000210\n"
        "cput r18, rfsl0\n"
        "get r18, rfsl0\n");

    dump(7);
    dump(8);
    dump(9);
    print("(b)\n");

#endif
    m = 0x91000000;
#define BLOCKS      20
#define WPB         4

    for(i = 0; i < (BLOCKS * WPB);i++) {
        m[i] = 0xaa000000U | i;
    }
    print("filled ");
    printhex(&m[0]);
    print(" to ");
    printhex(&m[(BLOCKS * WPB) - 1]);
    print(" with ");
    printhex(m[0]);
    print(" to ");
    printhex(m[(BLOCKS * WPB) - 1]);
    print("\n");

    for(i = 0; i < BLOCKS;i++) {
        unsigned k;
        j = &m[i * WPB];
        asm volatile(
            "put %2, rfsl0\n"
            "nop\nnop\n"
            "nop\nnop\n"
            "nop\nnop\n"
            "cput %1, rfsl0\n"
            "get %0, rfsl0\n" : "=r"(k) : "r"(j | 1), 
                        "r"((i % (BLOCKS / 2)) * 16));
        if ((i < 5) || (i == (BLOCKS - 1))) {
            //printhex(k);
            //print("\n");
        }
        if (i == 9) {
            print("initial value 9: ");
            dump(9);
        }
        if (i == 19) {
            print("new value 9: ");
            dump(9);
        }
    }
    print("pushed ");
    printhex(&m[0]);
    print(" to ");
    printhex(&m[(BLOCKS - 1) * WPB]);
    print(" with ");
    printhex(m[0]);
    print(" to ");
    printhex(m[(BLOCKS - 1) * WPB]);
    print("\n");

    print("\n");
    print("result of load at 8\n");
    dump(8);
    print("inspection of registers\n");
    inspection();
    print("\n");

    for(i = 0; i < BLOCKS;i++) {
        if ((i < 10) || (i > (BLOCKS - 3))) {
            print("pop ");
        }
        asm volatile(
            "addi r18, r0, (8*16)\n"
            "put r18, rfsl0\n"
            "nop\n"
            "nop\n"
            "nop\n"
            "nop\n"
            "cput r0, rfsl0\n"
            "get %0, rfsl0\n" : "=r"(j) );

        if ((i < 10) || (i > (BLOCKS - 3))) {
            printhex(j);
            print(" ");
            dump(8);
            inspection();
            print("\n");
        }
    }

    for(i = 0; i < 10;i++) {
        asm volatile(
            "put %1, rfsl0\n"
            "get %0, rfsl0\n" : "=r"(j) : "r"((i * 16)) );
        printhex(j);
        print("\n");
    }

    while(1) {}

    for (j = 0; j < 2000; j++) {
        unsigned a, b, c;
        FN ext_call;

        ext_mem_loc[0] = 0xb60f0008; // rtsd r15, 8
        ext_mem_loc[1] = 0x80000000; // nop

        asm volatile ("nop\nnop\nnop\nnop\n");
        asm volatile ("nop\nnop\nnop\nnop\n");
        asm volatile ("nop\nnop\nnop\nnop\n");
        ext_call = (FN) ext_mem_loc;

        a = get_time();
        ext_call();     // miss
        b = get_time();
        ext_call();     // hit
        c = get_time();
        c = (b - a) - (c - b);

        if (c > tmax) {
            tmax = c;
        }
        if (j == 10) {
            tmax = 0;
        }

        ext_mem_loc += 0x110;
    }
    print("cache timing - ");
    printhex(tmax);
    print("\n");

    while(1) {}
}


