
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity lmb is
    port (
        LMB_Clk                    : in std_logic := '0';
        LMB_Rst                    : in std_logic := '0';

        LMB_ABus                   : in  std_logic_vector(0 to 31);
        LMB_WriteDBus              : in  std_logic_vector(0 to 31);
        LMB_AddrStrobe             : in  std_logic;
        LMB_ReadStrobe             : in  std_logic;
        LMB_WriteStrobe            : in  std_logic;
        LMB_BE                     : in  std_logic_vector(0 to 3);
        LMB_DBus                   : out std_logic_vector(0 to 31);
        LMB_Ready                  : out std_logic;

        BRAM_Rst_A                 : out std_logic;
        BRAM_Clk_A                 : out std_logic;
        BRAM_Addr_A                : out std_logic_vector(0 to 31);
        BRAM_EN_A                  : out std_logic;
        BRAM_WEN_A                 : out std_logic_vector(0 to 3);
        BRAM_Dout_A                : out std_logic_vector(0 to 31);
        BRAM_Din_A                 : in  std_logic_vector(0 to 31);
        
        
        spm_addr                   : out std_logic_vector(31 downto 0);
        spm_to_cpu                 : in std_logic_vector(31 downto 0));
end lmb;


architecture SPM of lmb is

   signal lmb_select_0, lmb_select_1: std_logic;
   signal spm_0, lmb_select_d       : std_logic;

begin

    -- Address decoder
    lmb_select_0 <= '1' when LMB_ABus(0 to 3) = "0000" else '0';
    lmb_select_1 <= '1' when LMB_ABus(0 to 3) = "0001" else '0';

    -- Non-DMA (boot) SPM
    BRAM_Rst_A  <= '0';
    BRAM_Clk_A  <= LMB_Clk;
    BRAM_EN_A   <= '1';
    BRAM_WEN_A  <= LMB_BE 
                when (LMB_WriteStrobe and 
                        LMB_AddrStrobe and lmb_select_0) = '1' else "0000";
    BRAM_Dout_A <= LMB_WriteDBus;
    LMB_DBus    <= BRAM_Din_A when spm_0 = '1' else spm_to_cpu;
    BRAM_Addr_A <= LMB_ABus;

    Ready_Handling : process (LMB_Clk, LMB_Rst) is
    begin
        if (LMB_Rst = '1') then
            LMB_Ready <= '0';
            spm_0 <= '1';
            lmb_select_d <= '0';
        elsif (LMB_Clk'event and LMB_Clk = '1') then  -- rising clock edge
            LMB_Ready <= LMB_AddrStrobe and (lmb_select_1 or lmb_select_0);
            if (lmb_select_1 and LMB_AddrStrobe) = '1' then
                spm_0 <= '0';
            elsif (lmb_select_0 and LMB_AddrStrobe) = '1' then
                spm_0 <= '1';
            end if;
            lmb_select_d <= lmb_select_0;
        end if;
    end process Ready_Handling;

    spm_addr <= LMB_ABus;

end architecture SPM;


